#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Onionprobe test/monitor tool.
#
# Copyright (C) 2022 Silvio Rhatto <rhatto@torproject.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .config           import onionprobe_version
from prometheus_client import Counter, Gauge, Info, Enum

# Holds all metrics
metrics = {
    #
    # Metametrics: data about the Onionprobe instance itself
    #

    'onionprobe_version': Info(
        'onionprobe_version',
        'Onionprobe version information',
        ),

    'onionprobe_state': Enum(
        'onionprobe_state',
        'Onionprobe latest state',
        states=['starting', 'probing', 'sleeping', 'stopping']
        ),

    #
    # Probing gauges: the basic data
    #

    'onionprobe_wait': Gauge(
            'onionprobe_wait',
            'Records how long Onionprobe waited between two probes',
        ),

    'onion_service_latency': Gauge(
            'onion_service_latency',
            'Register Onion Service connection latency in seconds',
            ['name', 'address', 'protocol', 'port', 'path', 'updated_at']
        ),

    'onion_service_reachable': Gauge(
            'onion_service_reachable',
            'Register if the Onion Service is reachable: value is 1 for reachability and 0 otherwise',
            ['name', 'address', 'protocol', 'port', 'path', 'updated_at']
        ),

    'onion_service_connection_attempts': Gauge(
            'onion_service_connection_attempts',
            'Register the number of attempts when trying to connect to an Onion Service',
            ['name', 'address', 'protocol', 'port', 'path', 'reachable', 'updated_at']
        ),

    'onion_service_status_code': Gauge(
            'onion_service_status_code',
            'Register Onion Service connection HTTP status code',
            ['name', 'address', 'protocol', 'port', 'path', 'updated_at']
        ),

    'onion_service_descriptor_latency': Gauge(
            'onion_service_descriptor_latency',
            'Register Onion Service latency in seconds to get the descriptor',
            ['name', 'address', 'updated_at']
        ),

    'onion_service_descriptor_reachable': Gauge(
            'onion_service_descriptor_reachable',
            'Register if the Onion Service descriptor is available: value is 1 for reachability and 0 otherwise',
            ['name', 'address', 'hsdir', 'reason', 'updated_at']
        ),

    'onion_service_descriptor_fetch_attempts': Gauge(
            'onion_service_descriptor_fetch_attempts',
            'Register the number of attempts required when trying to get an Onion Service descriptor',
            ['name', 'address', 'reachable', 'updated_at']
        ),

    'onion_service_introduction_points_number': Gauge(
            'onion_service_introduction_points_number',
            'Register the number of introduction points in the Onion Service descriptor',
            ['name', 'address', 'updated_at']
        ),

    'onion_service_match_pattern_matched': Gauge(
            'onion_service_pattern_matched',
            'Register whether a regular expression pattern is matched when connection to the Onion Service: value is 1 for matched pattern and 0 otherwise',
            ['name', 'address', 'protocol', 'port', 'path', 'pattern', 'updated_at']
        ),

    'onion_service_valid_certificate': Gauge(
            'onion_service_valid_certificate',
            'Register whether the Onion Service HTTPS certificate is valid: value is 1 for valid and 0 otherwise, but only for sites reachable using HTTPS',
            ['name', 'address', 'protocol', 'port', 'path', 'updated_at']
        ),

    #
    # Probing counters
    #

    'onion_service_fetch_error_counter': Counter(
            'onion_service_fetch_error_counter',
            'Counts errors when fetching an Onion Service',
            ['name', 'address', 'protocol', 'port', 'path', 'updated_at']
        ),

    'onion_service_descriptor_fetch_error_counter': Counter(
            'onion_service_descriptor_fetch_error_counter',
            'Counts errors when fetching an Onion Service descriptor',
            ['name', 'address', 'updated_at']
        ),

    #
    # Requests exception counter
    #

    # Counter for requests.RequestException
    'onion_service_request_exception': Counter(
            'onion_service_request_exception',
            'Counts Onion Service general exception errors',
            ['name', 'address', 'protocol', 'port', 'path', 'updated_at']
        ),

    # Counter for requests.ConnectionError
    'onion_service_connection_error': Counter(
            'onion_service_connection_error',
            'Counts Onion Service connection errors',
            ['name', 'address', 'protocol', 'port', 'path', 'updated_at']
        ),

    # Counter for requests.HTTPError
    'onion_service_http_error': Counter(
            'onion_service_http_error',
            'Counts Onion Service HTTP errors',
            ['name', 'address', 'protocol', 'port', 'path', 'updated_at']
        ),

    # Counter for requests.TooManyRedirects
    'onion_service_too_many_redirects': Counter(
            'onion_service_too_many_redirects',
            'Counts Onion Service too many redirects errors',
            ['name', 'address', 'protocol', 'port', 'path', 'updated_at']
        ),

    # Counter for requests.ConnectionTimeout
    'onion_service_connection_timeout': Counter(
            'onion_service_connection_timeout',
            'Counts Onion Service connection timeouts',
            ['name', 'address', 'protocol', 'port', 'path', 'updated_at']
        ),

    # Counter for requests.ReadTimeout
    'onion_service_read_timeout': Counter(
            'onion_service_read_timeout',
            'Counts Onion Service read timeouts',
            ['name', 'address', 'protocol', 'port', 'path', 'updated_at']
        ),

    # Counter for requests.Timeout
    'onion_service_timeout': Counter(
            'onion_service_timeout',
            'Counts Onion Service timeouts',
            ['name', 'address', 'protocol', 'port', 'path', 'updated_at']
        ),

    # Counter for requests.exceptions.SSLError
    'onion_service_certificate_error': Counter(
            'onion_service_certificate_error',
            'Counts HTTPS certificate validation errors',
            ['name', 'address', 'protocol', 'port', 'path', 'updated_at']
        ),
    }

class OnionprobeMetrics:
    """
    Onionprobe class with metrics methods.
    """

    def initialize_prometheus_exporter(self):
        """
        Initialize the Prometheus Exporter
        """

        from prometheus_client import start_http_server

        port = self.get_config('prometheus_exporter_port')

        self.log('Initializing Prometheus HTTP exporter server at port %s...' % (port))
        start_http_server(port)

    def initialize_metrics(self):
        """
        Initialize the metrics subsystem

        It uses Prometheus metrics even if the Prometheus exporter is not in use.

        This means that the Prometheus metrics are always used, even if only for
        internal purposes, saving resources from preventing us to build additional
        metric logic.
        """

        # The metrics object
        self.metrics = metrics

        # Set version
        self.metrics['onionprobe_version'].info({
            'version': onionprobe_version,
            })

        # Set initial state
        self.metrics['onionprobe_state'].state('starting')

    def set_metric(self, metric, value, labels = {}):
        """
        Set a metric.

        :type  metric: str
        :param metric: Metric name

        :type  value: int
        :param value: Metric value

        :type  labels: dict
        :param labels: Metric labels dictionary.
                       Defaults to an empty dictionary.
        """

        if metric in self.metrics:
            # Apply timestamp
            labels['updated_at'] = self.timestamp()

            self.metrics[metric].labels(**labels).set(value)

    def inc_metric(self, metric, value = 1, labels = {}):
        """
        Increment a metric.

        :type  metric: str
        :param metric: Metric name

        :type  value: int
        :param value: Increment value. Defaults to 1.

        :type  labels: dict
        :param labels: Metric labels dictionary.
                       Defaults to an empty dictionary.
        """

        if metric in self.metrics:
            # Apply timestamp
            labels['updated_at'] = self.timestamp()

            self.metrics[metric].labels(**labels).inc(value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Onionprobe test/monitor tool.
#
# Copyright (C) 2022 Silvio Rhatto <rhatto@torproject.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .init       import OnionprobeInit
from .config     import OnionprobeConfig
from .logger     import OnionprobeLogger
from .time       import OnionprobeTime
from .tor        import OnionprobeTor
from .descriptor import OnionprobeDescriptor
from .metrics    import OnionprobeMetrics
from .prober     import OnionprobeProber
from .http       import OnionprobeHTTP
from .teardown   import OnionprobeTeardown
from .main       import OnionprobeMain

class Onionprobe(
        # Inherit from subsystems
        OnionprobeInit,
        OnionprobeConfig,
        OnionprobeLogger,
        OnionprobeTime,
        OnionprobeTor,
        OnionprobeDescriptor,
        OnionprobeMetrics,
        OnionprobeProber,
        OnionprobeHTTP,
        OnionprobeTeardown,
        OnionprobeMain,
        ):
    """
    Onionprobe class to test and monitor Tor Onion Services
    """

def run(args):
    """
    Run Onionprobe from arguments

    :type  args: dict
    :param args: Instance arguments.
    """

    # Dispatch
    try:
        probe = Onionprobe(args)

        if probe.initialize() is not False:
            probe.run()
            probe.close()
        else:
            print('Error: could not initialize')
            exit(1)

    #except (FileNotFoundError, KeyboardInterrupt) as e:
    except Exception as e:
        probe.log(e, 'error')
        probe.close()
        exit(1)

def run_from_cmdline():
    """
    Run Onionprobe getting arguments from the command line.
    """

    from .config import cmdline

    run(cmdline())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Handle Onionprobe configurations.
#
# Copyright (C) 2022 Silvio Rhatto <rhatto@torproject.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Dependencies
import os
import argparse

try:
    import yaml
except ImportError:
    print("Please install pyaml first!")
    raise ImportError

# The Onionprobe version string
# Uses Semantic Versioning 2.0.0
# See https://semver.org
onionprobe_version = '0.2.1'

# The base path for this project
basepath = os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir, os.pardir) + os.sep

# Default configuration
defaults = {
        'log_level'               : 'info',
        'launch_tor'              : True,
        'tor_address'             : '127.0.0.1',
        'socks_port'              : 19050,
        'control_port'            : 19051,
        'control_password'        : False,
        'loop'                    : False,
        'prometheus_exporter'     : False,
        'prometheus_exporter_port': 9091,
        'shuffle'                 : True,
        'randomize'               : True,
        'new_circuit'             : False,
        'interval'                : 60,
        'sleep'                   : 60,
        'descriptor_max_retries'  : 5,
        'descriptor_timeout'      : 30,
        'http_connect_timeout'    : 30,
        'http_connect_max_retries': 3,
        'http_read_timeout'       : 30,
        'circuit_stream_timeout'  : 60,
        'endpoints'               : {
            'www.torproject.org': {
                'address' : '2gzyxa5ihm7nsggfxnu52rck2vv4rvmdlkiu3zzui5du4xyclen53wid.onion',
                'protocol': 'http',
                'port'    : '80',
                'paths'   : [
                        {
                            'path'   : '/',
                            'pattern': 'Tor Project',
                        },
                    ],
                },
            }
        }

def cmdline():
    """
    Evalutate the command line.

    :return: Command line arguments.
    """

    epilog = """Examples:

      onionprobe -c configs/tor.yaml
      onionprobe -e http://2gzyxa5ihm7nsggfxnu52rck2vv4rvmdlkiu3zzui5du4xyclen53wid.onion
    """

    epilog += """\nAvailable metrics:
    """

    from .metrics import metrics

    for metric in metrics:
        item = metrics[metric].describe()[0]

        epilog += "\n  {}:\n        {}".format(item.name, item.documentation)

    description = 'Test and monitor onion services'
    parser      = argparse.ArgumentParser(
                    description=description,
                    epilog=epilog,
                    formatter_class=argparse.RawDescriptionHelpFormatter,
                  )

    parser.add_argument('-c', '--config', help='Read options from configuration file')

    parser.add_argument('-e', '--endpoints', nargs='*', help='Add endpoints to the test list', metavar="onion-address1")

    parser.add_argument('-v', '--version', action='version', version='%(prog)s ' + onionprobe_version)

    args = parser.parse_args()

    return args

class OnionprobeConfig:
    """
    Onionprobe class with configuration-related methods.
    """

    def get_config(self, item, default = None):
        """
        Helper to get instance configuration

        Retrieve a config parameter from the self.config object or use a
        default value as fallback

        :type  item: str
        :param item: Configuration item name

        :param default: Default config value to be used as a fallback if there's
                        no self.config[item] available.
                        Defaults to None

        :return: The configuration parameter value or the default fallback value.
        """

        if item in self.config:
            return self.config[item]

        # Optionally override the default with an argument provided
        elif default is not None:
            self.config[item] = default

            return default

        return defaults[item]

class OnionprobeConfigCompiler:
    """Base class to build Onionprobe configs from external sources of Onion Services"""

    def __init__(self, databases, template_config = None, output_path = None):
        """
        Constructor for the OnionprobeConfigCompiler class.

        Loads the default Onionprobe configuration to be used as a template.

        Keeps the dictionary of Onion Services databases as a class attribute.

        :type  databases: dict
        :param databases: Dictionary of data sources to fetch .onion sites.
                          Format is { 'database_name': 'database_url' }

        :type  template_config: str
        :param template_config: Configuration file path to be used as template

        :type  output_path: str
        :param output_path: Output folder where configs are written
        """

        # Save the databases of Onion Services
        self.databases = databases

        # Determine the default configuration file
        if template_config is None:
            template_config = os.path.join(basepath, 'configs', 'tor.yaml')

        # Determine the output path
        if output_path is None:
            self.output_path = os.path.join(basepath, 'configs')
        else:
            self.output_path = output_path

        # Load the default configuration file as a template
        if os.path.exists(template_config):
            with open(template_config, 'r') as config:
                self.config = yaml.load(config, yaml.CLoader)

    def build_endpoints_config(self, database):
        """
        Build the Onion Service endpoints dictionary.

        This method is only a placeholder.

        By default this method returns an empty dictionary as it's meant to be
        overriden by specific implementations inheriting from the
        OnionprobeConfigCompiler base class and where custom logic for
        extracting .onion endpoints from external databases should be located.

        :type database : str
        :param database: A database name from the databases dictionary. This
                         parameter allows accesing the URL of the external
                         database from the self.databases class attribute.

        :rtype: dict
        :return: Onion Service endpoints in the format accepted by Onionprobe.
        """

        return dict()

    def build_onionprobe_config(self):
        """
        Build an Onionprobe config.

        Writes an Onionprobe-compatible configuration file for each database
        listed in self.databases attribute.

        The Onion Service endpoints are generated from the
        build_endpoints_config() methods. To be effective, it's required that
        classes inheriting from this base class to implement the
        build_endpoints_configs() method.

        The filenames ared derived from the database names (each key from the
        self.databases attribute).
        """

        for database in self.databases:
            try:
                # Build list of endpoints
                endpoints = self.build_endpoints_config(database)

                # Create a new config using the default as base
                config = dict(self.config)

                # Replace the endpoints
                config['endpoints'] = endpoints

                # Save
                with open(os.path.join(self.output_path, database + '.yaml'), 'w') as output:
                    output.write(yaml.dump(config))

            except Exception as e:
                print(e)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Onionprobe test/monitor tool.
#
# Copyright (C) 2022 Silvio Rhatto <rhatto@torproject.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Dependencies
import logging

try:
    import stem.util
except ImportError:
    print("Please install stem library first!")
    raise ImportError

class OnionprobeLogger:
    """
    Onionprobe class with logging methods.
    """

    def initialize_logging(self):
        """
        Initialize Onionprobe's logging subsystem

        :rtype: bol
        :return: True if initialization is successful, False on error
        """

        log_level = self.get_config('log_level').upper()

        if log_level in dir(logging):
            level = getattr(logging, log_level)

            logging.basicConfig(level=level)

            # See https://stem.torproject.org/api/util/log.html
            stem_logger = stem.util.log.get_logger()

            stem_logger.setLevel(level)

        else:
            logging.error("Invalid log level %s" % (log_level))

            return False

        return True

    def log(self, message, level='info'):
        """
        Helper log function

        Appends a message into the logging subsystem.

        :type  message: str
        :param message: The message to be logged.

        :type  level: str
        :param level: The log level. Defaults to 'info'.
                      For the available log levels, check
                      https://docs.python.org/3/howto/logging.html#logging-levels
        """

        # Just a wrapper for the logging() function
        getattr(logging, level)(message)
